@@ -72,7 +72,7 @@ const (
 	REG_X30
 	REG_X31
 
-	// FP register numberings.
+	// Floating Point register numberings.
 	REG_F0
 	REG_F1
 	REG_F2
@@ -106,6 +106,40 @@ const (
 	REG_F30
 	REG_F31
 
+	// Vector register numberings.
+	REG_V0
+	REG_V1
+	REG_V2
+	REG_V3
+	REG_V4
+	REG_V5
+	REG_V6
+	REG_V7
+	REG_V8
+	REG_V9
+	REG_V10
+	REG_V11
+	REG_V12
+	REG_V13
+	REG_V14
+	REG_V15
+	REG_V16
+	REG_V17
+	REG_V18
+	REG_V19
+	REG_V20
+	REG_V21
+	REG_V22
+	REG_V23
+	REG_V24
+	REG_V25
+	REG_V26
+	REG_V27
+	REG_V28
+	REG_V29
+	REG_V30
+	REG_V31
+
 	// This marks the end of the register numbering.
 	REG_END
 
@@ -290,11 +324,13 @@ const (
 //
 // As well as some pseudo-mnemonics (e.g. MOV) used only in the assembler.
 //
-// See also "The RISC-V Instruction Set Manual" at https://riscv.org/specifications/.
+// See also "The RISC-V Instruction Set Manual" at https://riscv.org/technical/specifications/.
 //
 // If you modify this table, you MUST run 'go generate' to regenerate anames.go!
 const (
-	// Unprivileged ISA (Document Version 20190608-Base-Ratified)
+	//
+	// Unprivileged ISA (version 20240411)
+	//
 
 	// 2.4: Integer Computational Instructions
 	AADDI = obj.ABaseRISCV + obj.A_ARCHSPECIFIC + iota
@@ -342,10 +378,8 @@ const (
 
 	// 2.7: Memory Ordering Instructions
 	AFENCE
-	AFENCETSO
-	APAUSE
 
-	// 5.2: Integer Computational Instructions (RV64I)
+	// 4.2: Integer Computational Instructions (RV64I)
 	AADDIW
 	ASLLIW
 	ASRLIW
@@ -356,16 +390,26 @@ const (
 	ASUBW
 	ASRAW
 
-	// 5.3: Load and Store Instructions (RV64I)
+	// 4.3: Load and Store Instructions (RV64I)
 	ALD
 	ASD
 
-	// 7.1: Multiplication Operations
+	// 7.1: CSR Instructions (Zicsr)
+	ACSRRW
+	ACSRRS
+	ACSRRC
+	ACSRRWI
+	ACSRRSI
+	ACSRRCI
+
+	// 13.1: Multiplication Operations
 	AMUL
 	AMULH
 	AMULHU
 	AMULHSU
 	AMULW
+
+	// 13.2: Division Operations
 	ADIV
 	ADIVU
 	AREM
@@ -375,13 +419,13 @@ const (
 	AREMW
 	AREMUW
 
-	// 8.2: Load-Reserved/Store-Conditional Instructions
+	// 14.2: Load-Reserved/Store-Conditional Instructions (Zalrsc)
 	ALRD
 	ASCD
 	ALRW
 	ASCW
 
-	// 8.3: Atomic Memory Operations
+	// 14.4: Atomic Memory Operations (Zaamo)
 	AAMOSWAPD
 	AAMOADDD
 	AAMOANDD
@@ -401,29 +445,11 @@ const (
 	AAMOMINW
 	AAMOMINUW
 
-	// 10.1: Base Counters and Timers
-	ARDCYCLE
-	ARDCYCLEH
-	ARDTIME
-	ARDTIMEH
-	ARDINSTRET
-	ARDINSTRETH
-
-	// 11.2: Floating-Point Control and Status Register
-	AFRCSR
-	AFSCSR
-	AFRRM
-	AFSRM
-	AFRFLAGS
-	AFSFLAGS
-	AFSRMI
-	AFSFLAGSI
-
-	// 11.5: Single-Precision Load and Store Instructions
+	// 20.5: Single-Precision Load and Store Instructions
 	AFLW
 	AFSW
 
-	// 11.6: Single-Precision Floating-Point Computational Instructions
+	// 20.6: Single-Precision Floating-Point Computational Instructions
 	AFADDS
 	AFSUBS
 	AFMULS
@@ -436,7 +462,7 @@ const (
 	AFNMADDS
 	AFNMSUBS
 
-	// 11.7: Single-Precision Floating-Point Conversion and Move Instructions
+	// 20.7: Single-Precision Floating-Point Conversion and Move Instructions
 	AFCVTWS
 	AFCVTLS
 	AFCVTSW
@@ -453,19 +479,19 @@ const (
 	AFMVXW
 	AFMVWX
 
-	// 11.8: Single-Precision Floating-Point Compare Instructions
+	// 20.8: Single-Precision Floating-Point Compare Instructions
 	AFEQS
 	AFLTS
 	AFLES
 
-	// 11.9: Single-Precision Floating-Point Classify Instruction
+	// 20.9: Single-Precision Floating-Point Classify Instruction
 	AFCLASSS
 
-	// 12.3: Double-Precision Load and Store Instructions
+	// 21.3: Double-Precision Load and Store Instructions
 	AFLD
 	AFSD
 
-	// 12.4: Double-Precision Floating-Point Computational Instructions
+	// 21.4: Double-Precision Floating-Point Computational Instructions
 	AFADDD
 	AFSUBD
 	AFMULD
@@ -478,7 +504,7 @@ const (
 	AFNMADDD
 	AFNMSUBD
 
-	// 12.5: Double-Precision Floating-Point Conversion and Move Instructions
+	// 21.5: Double-Precision Floating-Point Conversion and Move Instructions
 	AFCVTWD
 	AFCVTLD
 	AFCVTDW
@@ -495,19 +521,19 @@ const (
 	AFMVXD
 	AFMVDX
 
-	// 12.6: Double-Precision Floating-Point Compare Instructions
+	// 21.6: Double-Precision Floating-Point Compare Instructions
 	AFEQD
 	AFLTD
 	AFLED
 
-	// 12.7: Double-Precision Floating-Point Classify Instruction
+	// 21.7: Double-Precision Floating-Point Classify Instruction
 	AFCLASSD
 
-	// 13.1 Quad-Precision Load and Store Instructions
+	// 22.1 Quad-Precision Load and Store Instructions
 	AFLQ
 	AFSQ
 
-	// 13.2: Quad-Precision Computational Instructions
+	// 22.2: Quad-Precision Computational Instructions
 	AFADDQ
 	AFSUBQ
 	AFMULQ
@@ -520,7 +546,7 @@ const (
 	AFNMADDQ
 	AFNMSUBQ
 
-	// 13.3 Quad-Precision Convert and Move Instructions
+	// 22.3 Quad-Precision Convert and Move Instructions
 	AFCVTWQ
 	AFCVTLQ
 	AFCVTSQ
@@ -537,46 +563,15 @@ const (
 	AFSGNJNQ
 	AFSGNJXQ
 
-	// 13.4 Quad-Precision Floating-Point Compare Instructions
+	// 22.4 Quad-Precision Floating-Point Compare Instructions
 	AFEQQ
 	AFLEQ
 	AFLTQ
 
-	// 13.5 Quad-Precision Floating-Point Classify Instruction
+	// 22.5 Quad-Precision Floating-Point Classify Instruction
 	AFCLASSQ
 
-	// Privileged ISA (Version 20190608-Priv-MSU-Ratified)
-
-	// 3.1.9: Instructions to Access CSRs
-	ACSRRW
-	ACSRRS
-	ACSRRC
-	ACSRRWI
-	ACSRRSI
-	ACSRRCI
-
-	// 3.2.1: Environment Call and Breakpoint
-	AECALL
-	ASCALL
-	AEBREAK
-	ASBREAK
-
-	// 3.2.2: Trap-Return Instructions
-	AMRET
-	ASRET
-	ADRET
-
-	// 3.2.3: Wait for Interrupt
-	AWFI
-
-	// 4.2.1: Supervisor Memory-Management Fence Instruction
-	ASFENCEVMA
-
-	//
-	// RISC-V Bit-Manipulation ISA-extensions (1.0)
-	//
-
-	// 1.1: Address Generation Instructions (Zba)
+	// 28.4.1: Address Generation Instructions (Zba)
 	AADDUW
 	ASH1ADD
 	ASH1ADDUW
@@ -586,7 +581,7 @@ const (
 	ASH3ADDUW
 	ASLLIUW
 
-	// 1.2: Basic Bit Manipulation (Zbb)
+	// 28.4.2: Basic Bit Manipulation (Zbb)
 	AANDN
 	AORN
 	AXNOR
@@ -604,7 +599,7 @@ const (
 	ASEXTH
 	AZEXTH
 
-	// 1.3: Bitwise Rotation (Zbb)
+	// 28.4.3: Bitwise Rotation (Zbb)
 	AROL
 	AROLW
 	AROR
@@ -614,7 +609,7 @@ const (
 	AORCB
 	AREV8
 
-	// 1.5: Single-bit Instructions (Zbs)
+	// 28.4.4: Single-bit Instructions (Zbs)
 	ABCLR
 	ABCLRI
 	ABEXT
@@ -624,6 +619,518 @@ const (
 	ABSET
 	ABSETI
 
+	//
+	// RISC-V Vector ISA-extension (1.0) (Unprivileged 20240411)
+	//
+
+	// 31.6. Configuration-Setting Instructions
+	AVSETVLI
+	AVSETIVLI
+	AVSETVL
+
+	// 31.7.4. Vector Unit-Stride Instructions
+	AVLE8V
+	AVLE16V
+	AVLE32V
+	AVLE64V
+	AVSE8V
+	AVSE16V
+	AVSE32V
+	AVSE64V
+	AVLMV
+	AVSMV
+
+	// 31.7.5. Vector Strided Instructions
+	AVLSE8V
+	AVLSE16V
+	AVLSE32V
+	AVLSE64V
+	AVSSE8V
+	AVSSE16V
+	AVSSE32V
+	AVSSE64V
+
+	// 31.7.6. Vector Indexed Instructions
+	AVLUXEI8V
+	AVLUXEI16V
+	AVLUXEI32V
+	AVLUXEI64V
+	AVLOXEI8V
+	AVLOXEI16V
+	AVLOXEI32V
+	AVLOXEI64V
+	AVSUXEI8V
+	AVSUXEI16V
+	AVSUXEI32V
+	AVSUXEI64V
+	AVSOXEI8V
+	AVSOXEI16V
+	AVSOXEI32V
+	AVSOXEI64V
+
+	// 31.7.7. Unit-stride Fault-Only-First Loads
+	AVLE8FFV
+	AVLE16FFV
+	AVLE32FFV
+	AVLE64FFV
+
+	// 31.7.9. Vector Load/Store Whole Register Instructions
+	AVL1RE8V
+	AVL1RE16V
+	AVL1RE32V
+	AVL1RE64V
+	AVL2RE8V
+	AVL2RE16V
+	AVL2RE32V
+	AVL2RE64V
+	AVL4RE8V
+	AVL4RE16V
+	AVL4RE32V
+	AVL4RE64V
+	AVL8RE8V
+	AVL8RE16V
+	AVL8RE32V
+	AVL8RE64V
+	AVS1RV
+	AVS2RV
+	AVS4RV
+	AVS8RV
+
+	// 31.11.1. Vector Single-Width Integer Add and Subtract
+	AVADDVV
+	AVADDVX
+	AVADDVI
+	AVSUBVV
+	AVSUBVX
+	AVRSUBVX
+	AVRSUBVI
+
+	// 31.11.2. Vector Widening Integer Add/Subtract
+	AVWADDUVV
+	AVWADDUVX
+	AVWSUBUVV
+	AVWSUBUVX
+	AVWADDVV
+	AVWADDVX
+	AVWSUBVV
+	AVWSUBVX
+	AVWADDUWV
+	AVWADDUWX
+	AVWSUBUWV
+	AVWSUBUWX
+	AVWADDWV
+	AVWADDWX
+	AVWSUBWV
+	AVWSUBWX
+
+	// 31.11.3. Vector Integer Extension
+	AVZEXTVF2
+	AVSEXTVF2
+	AVZEXTVF4
+	AVSEXTVF4
+	AVZEXTVF8
+	AVSEXTVF8
+
+	// 31.11.4. Vector Integer Add-with-Carry / Subtract-with-Borrow Instructions
+	AVADCVVM
+	AVADCVXM
+	AVADCVIM
+	AVMADCVVM
+	AVMADCVXM
+	AVMADCVIM
+	AVMADCVV
+	AVMADCVX
+	AVMADCVI
+	AVSBCVVM
+	AVSBCVXM
+	AVMSBCVVM
+	AVMSBCVXM
+	AVMSBCVV
+	AVMSBCVX
+
+	// 31.11.5. Vector Bitwise Logical Instructions
+	AVANDVV
+	AVANDVX
+	AVANDVI
+	AVORVV
+	AVORVX
+	AVORVI
+	AVXORVV
+	AVXORVX
+	AVXORVI
+
+	// 31.11.6. Vector Single-Width Shift Instructions
+	AVSLLVV
+	AVSLLVX
+	AVSLLVI
+	AVSRLVV
+	AVSRLVX
+	AVSRLVI
+	AVSRAVV
+	AVSRAVX
+	AVSRAVI
+
+	// 31.11.7. Vector Narrowing Integer Right Shift Instructions
+	AVNSRLWV
+	AVNSRLWX
+	AVNSRLWI
+	AVNSRAWV
+	AVNSRAWX
+	AVNSRAWI
+
+	// 31.11.8. Vector Integer Compare Instructions
+	AVMSEQVV
+	AVMSEQVX
+	AVMSEQVI
+	AVMSNEVV
+	AVMSNEVX
+	AVMSNEVI
+	AVMSLTUVV
+	AVMSLTUVX
+	AVMSLTVV
+	AVMSLTVX
+	AVMSLEUVV
+	AVMSLEUVX
+	AVMSLEUVI
+	AVMSLEVV
+	AVMSLEVX
+	AVMSLEVI
+	AVMSGTUVX
+	AVMSGTUVI
+	AVMSGTVX
+	AVMSGTVI
+
+	// 31.11.9. Vector Integer Min/Max Instructions
+	AVMINUVV
+	AVMINUVX
+	AVMINVV
+	AVMINVX
+	AVMAXUVV
+	AVMAXUVX
+	AVMAXVV
+	AVMAXVX
+
+	// 31.11.10. Vector Single-Width Integer Multiply Instructions
+	AVMULVV
+	AVMULVX
+	AVMULHVV
+	AVMULHVX
+	AVMULHUVV
+	AVMULHUVX
+	AVMULHSUVV
+	AVMULHSUVX
+
+	// 31.11.11. Vector Integer Divide Instructions
+	AVDIVUVV
+	AVDIVUVX
+	AVDIVVV
+	AVDIVVX
+	AVREMUVV
+	AVREMUVX
+	AVREMVV
+	AVREMVX
+
+	// 31.11.12. Vector Widening Integer Multiply Instructions
+	AVWMULVV
+	AVWMULVX
+	AVWMULUVV
+	AVWMULUVX
+	AVWMULSUVV
+	AVWMULSUVX
+
+	// 31.11.13. Vector Single-Width Integer Multiply-Add Instructions
+	AVMACCVV
+	AVMACCVX
+	AVNMSACVV
+	AVNMSACVX
+	AVMADDVV
+	AVMADDVX
+	AVNMSUBVV
+	AVNMSUBVX
+
+	// 31.11.14. Vector Widening Integer Multiply-Add Instructions
+	AVWMACCUVV
+	AVWMACCUVX
+	AVWMACCVV
+	AVWMACCVX
+	AVWMACCSUVV
+	AVWMACCSUVX
+	AVWMACCUSVX
+
+	// 31.11.15. Vector Integer Merge Instructions
+	AVMERGEVVM
+	AVMERGEVXM
+	AVMERGEVIM
+
+	// 31.11.16. Vector Integer Move Instructions
+	AVMVVV
+	AVMVVX
+	AVMVVI
+
+	// 31.12.1. Vector Single-Width Saturating Add and Subtract
+	AVSADDUVV
+	AVSADDUVX
+	AVSADDUVI
+	AVSADDVV
+	AVSADDVX
+	AVSADDVI
+	AVSSUBUVV
+	AVSSUBUVX
+	AVSSUBVV
+	AVSSUBVX
+
+	// 31.12.2. Vector Single-Width Averaging Add and Subtract
+	AVAADDUVV
+	AVAADDUVX
+	AVAADDVV
+	AVAADDVX
+	AVASUBUVV
+	AVASUBUVX
+	AVASUBVV
+	AVASUBVX
+
+	// 31.12.3. Vector Single-Width Fractional Multiply with Rounding and Saturation
+	AVSMULVV
+	AVSMULVX
+
+	// 31.12.4. Vector Single-Width Scaling Shift Instructions
+	AVSSRLVV
+	AVSSRLVX
+	AVSSRLVI
+	AVSSRAVV
+	AVSSRAVX
+	AVSSRAVI
+
+	// 31.12.5. Vector Narrowing Fixed-Point Clip Instructions
+	AVNCLIPUWV
+	AVNCLIPUWX
+	AVNCLIPUWI
+	AVNCLIPWV
+	AVNCLIPWX
+	AVNCLIPWI
+
+	// 31.13.2. Vector Single-Width Floating-Point Add/Subtract Instructions
+	AVFADDVV
+	AVFADDVF
+	AVFSUBVV
+	AVFSUBVF
+	AVFRSUBVF
+
+	// 31.13.3. Vector Widening Floating-Point Add/Subtract Instructions
+	AVFWADDVV
+	AVFWADDVF
+	AVFWSUBVV
+	AVFWSUBVF
+	AVFWADDWV
+	AVFWADDWF
+	AVFWSUBWV
+	AVFWSUBWF
+
+	// 31.13.4. Vector Single-Width Floating-Point Multiply/Divide Instructions
+	AVFMULVV
+	AVFMULVF
+	AVFDIVVV
+	AVFDIVVF
+	AVFRDIVVF
+
+	// 31.13.5. Vector Widening Floating-Point Multiply
+	AVFWMULVV
+	AVFWMULVF
+
+	// 31.13.6. Vector Single-Width Floating-Point Fused Multiply-Add Instructions
+	AVFMACCVV
+	AVFMACCVF
+	AVFNMACCVV
+	AVFNMACCVF
+	AVFMSACVV
+	AVFMSACVF
+	AVFNMSACVV
+	AVFNMSACVF
+	AVFMADDVV
+	AVFMADDVF
+	AVFNMADDVV
+	AVFNMADDVF
+	AVFMSUBVV
+	AVFMSUBVF
+	AVFNMSUBVV
+	AVFNMSUBVF
+
+	// 31.13.7. Vector Widening Floating-Point Fused Multiply-Add Instructions
+	AVFWMACCVV
+	AVFWMACCVF
+	AVFWNMACCVV
+	AVFWNMACCVF
+	AVFWMSACVV
+	AVFWMSACVF
+	AVFWNMSACVV
+	AVFWNMSACVF
+
+	// 31.13.8. Vector Floating-Point Square-Root Instruction
+	AVFSQRTV
+
+	// 31.13.9. Vector Floating-Point Reciprocal Square-Root Estimate Instruction
+	AVFRSQRT7V
+
+	// 31.13.10. Vector Floating-Point Reciprocal Estimate Instruction
+	AVFREC7V
+
+	// 31.13.11. Vector Floating-Point MIN/MAX Instructions
+	AVFMINVV
+	AVFMINVF
+	AVFMAXVV
+	AVFMAXVF
+
+	// 31.13.12. Vector Floating-Point Sign-Injection Instructions
+	AVFSGNJVV
+	AVFSGNJVF
+	AVFSGNJNVV
+	AVFSGNJNVF
+	AVFSGNJXVV
+	AVFSGNJXVF
+
+	// 31.13.13. Vector Floating-Point Compare Instructions
+	AVMFEQVV
+	AVMFEQVF
+	AVMFNEVV
+	AVMFNEVF
+	AVMFLTVV
+	AVMFLTVF
+	AVMFLEVV
+	AVMFLEVF
+	AVMFGTVF
+	AVMFGEVF
+
+	// 31.13.14. Vector Floating-Point Classify Instruction
+	AVFCLASSV
+
+	// 31.13.15. Vector Floating-Point Merge Instruction
+	AVFMERGEVFM
+
+	// 31.13.16. Vector Floating-Point Move Instruction
+	AVFMVVF
+
+	// 31.13.17. Single-Width Floating-Point/Integer Type-Convert Instructions
+	AVFCVTXUFV
+	AVFCVTXFV
+	AVFCVTRTZXUFV
+	AVFCVTRTZXFV
+	AVFCVTFXUV
+	AVFCVTFXV
+
+	// 31.13.18. Widening Floating-Point/Integer Type-Convert Instructions
+	AVFWCVTXUFV
+	AVFWCVTXFV
+	AVFWCVTRTZXUFV
+	AVFWCVTRTZXFV
+	AVFWCVTFXUV
+	AVFWCVTFXV
+	AVFWCVTFFV
+
+	// 31.13.19. Narrowing Floating-Point/Integer Type-Convert Instructions
+	AVFNCVTXUFW
+	AVFNCVTXFW
+	AVFNCVTRTZXUFW
+	AVFNCVTRTZXFW
+	AVFNCVTFXUW
+	AVFNCVTFXW
+	AVFNCVTFFW
+	AVFNCVTRODFFW
+
+	// 31.14.1. Vector Single-Width Integer Reduction Instructions
+	AVREDSUMVS
+	AVREDMAXUVS
+	AVREDMAXVS
+	AVREDMINUVS
+	AVREDMINVS
+	AVREDANDVS
+	AVREDORVS
+	AVREDXORVS
+
+	// 31.14.2. Vector Widening Integer Reduction Instructions
+	AVWREDSUMUVS
+	AVWREDSUMVS
+
+	// 31.14.3. Vector Single-Width Floating-Point Reduction Instructions
+	AVFREDOSUMVS
+	AVFREDUSUMVS
+	AVFREDMAXVS
+	AVFREDMINVS
+
+	// 31.14.4. Vector Widening Floating-Point Reduction Instructions
+	AVFWREDOSUMVS
+	AVFWREDUSUMVS
+
+	// 31.15. Vector Mask Instructions
+	AVMANDMM
+	AVMNANDMM
+	AVMANDNMM
+	AVMXORMM
+	AVMORMM
+	AVMNORMM
+	AVMORNMM
+	AVMXNORMM
+	AVCPOPM
+	AVFIRSTM
+	AVMSBFM
+	AVMSIFM
+	AVMSOFM
+	AVIOTAM
+	AVIDV
+
+	// 31.16.1. Integer Scalar Move Instructions
+	AVMVXS
+	AVMVSX
+
+	// 31.16.2. Floating-Point Scalar Move Instructions
+	AVFMVFS
+	AVFMVSF
+
+	// 31.16.3. Vector Slide Instructions
+	AVSLIDEUPVX
+	AVSLIDEUPVI
+	AVSLIDEDOWNVX
+	AVSLIDEDOWNVI
+	AVSLIDE1UPVX
+	AVFSLIDE1UPVF
+	AVSLIDE1DOWNVX
+	AVFSLIDE1DOWNVF
+
+	// 31.16.4. Vector Register Gather Instructions
+	AVRGATHERVV
+	AVRGATHEREI16VV
+	AVRGATHERVX
+	AVRGATHERVI
+
+	// 31.16.5. Vector Compress Instruction
+	AVCOMPRESSVM
+
+	// 31.16.6. Whole Vector Register Move
+	AVMV1RV
+	AVMV2RV
+	AVMV4RV
+	AVMV8RV
+
+	//
+	// Privileged ISA (version 20240411)
+	//
+
+	// 3.3.1: Environment Call and Breakpoint
+	AECALL
+	ASCALL
+	AEBREAK
+	ASBREAK
+
+	// 3.3.2: Trap-Return Instructions
+	AMRET
+	ASRET
+	ADRET
+
+	// 3.3.3: Wait for Interrupt
+	AWFI
+
+	// 10.2: Supervisor Memory-Management Fence Instruction
+	ASFENCEVMA
+
 	// The escape hatch. Inserts a single 32-bit word.
 	AWORD
 
@@ -641,15 +1148,15 @@ const (
 	ABNEZ
 	AFABSD
 	AFABSS
+	AFNED
 	AFNEGD
 	AFNEGS
-	AFNED
 	AFNES
 	AMOV
 	AMOVB
 	AMOVBU
-	AMOVF
 	AMOVD
+	AMOVF
 	AMOVH
 	AMOVHU
 	AMOVW
@@ -657,6 +1164,9 @@ const (
 	ANEG
 	ANEGW
 	ANOT
+	ARDCYCLE
+	ARDINSTRET
+	ARDTIME
 	ASEQZ
 	ASNEZ
 
@@ -715,12 +1225,9 @@ const (
 // Any instructions not listed here are assumed to either be non-unary or to read
 // from its argument.
 var unaryDst = map[obj.As]bool{
-	ARDCYCLE:    true,
-	ARDCYCLEH:   true,
-	ARDTIME:     true,
-	ARDTIMEH:    true,
-	ARDINSTRET:  true,
-	ARDINSTRETH: true,
+	ARDCYCLE:   true,
+	ARDTIME:    true,
+	ARDINSTRET: true,
 }
 
 // Instruction encoding masks.
