@@ -84,12 +84,17 @@ const (
 
 	_EV_ADD       = 0x1
 	_EV_DELETE    = 0x2
+	_EV_ENABLE    = 0x4
+	_EV_DISABLE   = 0x8
 	_EV_CLEAR     = 0x20
 	_EV_RECEIPT   = 0
 	_EV_ERROR     = 0x4000
 	_EV_EOF       = 0x8000
 	_EVFILT_READ  = 0x0
 	_EVFILT_WRITE = 0x1
+	_EVFILT_USER  = 0x8
+
+	_NOTE_TRIGGER = 0x1000000
 )
 
 type sigset struct {
